(*
  TRS representation of HOL equations.
  Basic tactics
*)

signature ISAFOR_BASICS =
sig
  include ISAFOR_TERM

  val trsT : typ
  val ruleT : typ
  val trs_repS : sort
  val embed_const : typ -> term
  val union_const_name : string

  val mk_embed : term -> term
  val mk_rstep : term -> term
  val mk_star : term -> term
  val maybe_mk_embed : theory -> term -> term

  val dest_embed : term -> term

  val mk_inrel : (term * term) -> term -> term
  val dest_inrel : term -> ((term * term) * term)

  val args_rsteps_tac : (int -> tactic) -> int -> tactic 
  val supteq_tac : int list -> int -> tactic

  val subst_closure : Proof.context -> thm -> thm

  val restrict_sortsT : typ -> typ
  val restrict_sorts : term -> term
end

structure Isafor_Basics : ISAFOR_BASICS =
struct

open Isafor_Term


(* types and constants *)

val ruleT = HOLogic.mk_prodT (termT, termT)
val trsT = HOLogic.mk_setT ruleT
val trs_repS = @{sort TermFun_Setup.trs_rep}

val union_const_name = @{const_name Lattices.semilattice_sup_class.sup}

fun embed_const T = Const (@{const_name TermFun_Setup.trs_rep_class.embed}, T --> termT)

val rstep_const = Const (@{const_name Trs.rstep}, trsT --> trsT)
val rtrancl_const = Const (@{const_name Transitive_Closure.rtrancl}, trsT --> trsT)


(* building TRSs from definitions *)

fun mk_embed t = embed_const (fastype_of t) $ t

fun maybe_mk_embed thy t =
  let 
    val T = fastype_of t
  in 
    if Sign.of_sort thy (T, trs_repS) then mk_embed t else dummy_term
  end

fun mk_rstep t = rstep_const $ t

fun mk_star t = rtrancl_const $ t


(* destruction *)

fun dest_embed (Const (@{const_name TermFun_Setup.trs_rep_class.embed}, _) $ t) = t

fun mk_inrel (x, y) R =
  HOLogic.mk_Trueprop (HOLogic.mk_mem (HOLogic.mk_prod (x, y), R))

fun dest_inrel t =
  let val (pair, rel) = HOLogic.dest_mem (HOLogic.dest_Trueprop t)
  in (HOLogic.dest_prod pair, rel) end

fun args_rsteps_tac sub_tac = SUBGOAL (fn (t,i) =>
  let 
    val concl = Logic.strip_assums_concl t
    val ((l,_), _) = dest_inrel concl
    val (_, ts) = dest_app l
    val arity = length ts
  in
    rtac @{thm Trs.args_rsteps_imp_rsteps} i
    THEN REPEAT_DETERM_N arity (rtac @{thm TermFun_Setup.length_eqI} i)
    THEN rtac @{thm HOL.refl} i
    THEN EVERY (map (rtac @{thm TermFun_Setup.pairwiseI2}) (i upto i + arity - 1))
    THEN rtac @{thm TermFun_Setup.pairwiseI1} (i + arity)
    THEN EVERY (map sub_tac (i + arity - 1 downto i))
  end)


fun subst_closure ctxt thm =
  let
    val thy = ProofContext.theory_of ctxt
    val cstringT = ctyp_of thy HOLogic.stringT
    val ((l, r), R) = dest_inrel (prop_of thm)
    val sigma = 
      union (op =) (vars_of l) (vars_of r)
      |> map (fn v => (HOLogic.mk_string v, Var ((v, 0), termT)))
      |> mk_subst
    
    val cert = SOME o cterm_of thy 
    val inst = Drule.instantiate' [SOME cstringT, SOME cstringT] (map cert [l, r, R, sigma])
  in
    Thm.implies_elim (inst @{thm Trs.subst_closure.subst_closureI}) thm
    |> Simplifier.full_simplify (HOL_basic_ss addsimps
        @{thms Term.subst_apply_term.simps
               List.map.simps (*prefix important, due to RBT_Impl.map.simps*)
               Fun.fun_upd_apply
               HOL.if_True
               HOL.if_False
               List.list.inject
               List.list.distinct
               String.char.inject
               (*String.char.distinct (*does not exist*)*)
               (*nibble.inject (*does not exist*)*)
               String.nibble.eq.simps
               String.nibble.distinct
               HOL.simp_thms(21-24)
               HOL.refl
               })
  end

fun list_member_tac m i =
  if m = 0 then rtac @{thm TermFun_Setup.list_memberI1} i
  else rtac @{thm TermFun_Setup.list_memberI2} i THEN list_member_tac (m - 1) i

fun supteq_tac [] i = rtac @{thm Term.supteq.refl} i
  | supteq_tac (j::js) i = 
    rtac @{thm Term.supteq.subt} i
    THEN list_member_tac j i
    THEN supteq_tac js i


(* add type class constraint "trs_rep" to all type vars in term *)
local
  fun restrict (TVar (ixn, S)) = TVar (ixn, @{sort TermFun_Setup.trs_rep} @ S)
    | restrict (TFree (n, S)) = TFree (n, @{sort TermFun_Setup.trs_rep} @ S)
in
  val restrict_sortsT = map_atyps restrict
  val restrict_sorts = map_types restrict_sortsT
end


end
